#!/usr/bin/env python
from __future__ import print_function


"""pcap-player.py: replay tcpdump/wireshark captures"""
__version__   = "1.0"
__author__    = "ed <pcapplayer@ocv.me>"
__credits__   = ["stackoverflow.com"]
__url__       = "https://ocv.me/dev/?pcap-player.py"
__license__   = "MIT"
__copyright__ = 2017


import sys
import os
import time
import gzip
import socket
import struct
import threading
import argparse
from pprint import pprint
from datetime import datetime


sys.path.insert(1, '.')
try:
	import dpkt
except ImportError as e:
	print('\n  need the dpkt library; either installed globally or located directly inside a folder next to this script.\n  see https://github.com/kbandla/dpkt/releases (tested with 1.9.1)\n')
	raise e


t_real = None
paused = None
dropping = None
targets = {}
move_up = 0


if sys.platform.startswith('linux') \
or sys.platform == 'darwin':
	import tty, termios
	def getch():
		ch = 0
		fd = sys.stdin.fileno()
		old_cfg = termios.tcgetattr(fd)
		tmp_cfg = termios.tcgetattr(fd)
		tmp_cfg[3] &= ~termios.ICANON  # & ~termios.ECHO
		try:
			#tty.setraw(fd)
			termios.tcsetattr(fd, termios.TCSADRAIN, tmp_cfg)
			ch = sys.stdin.read(1)
			sys.stdout.flush()
		finally:
			termios.tcsetattr(fd, termios.TCSADRAIN, old_cfg)
		return ch

elif sys.platform in ['win32','cygwin']:
	import msvcrt
	def getch():
		while msvcrt.kbhit():
			msvcrt.getch()
		
		ch = msvcrt.getch()
		if sys.version_info[0] > 2:
			return ch
		else:
			return ch.decode(encoding='mbcs')

else:
	raise RuntimeError('unsupported platform: {0}'.format(sys.platform))


def keyboard():
	global paused, dropping, t_real, move_up
	while True:
		cmd = getch()
		
		if cmd == 'p':
			if paused is None:
				paused = time.time()
				print('\033[D** PAUSED **')
				move_up += 1
			else:
				now = time.time()
				t_real += (now - paused)
				paused = None
		
		if cmd == 'd':
			if dropping is None:
				dropping = True
			else:
				dropping = None
		
		if cmd == 's':
			dropping = 1


def get_packets(pcap):
	for ts, pkt in pcap:
		eth = dpkt.ethernet.Ethernet(pkt)
		ip = eth.data
		
		if eth.type == dpkt.ethernet.ETH_TYPE_IP \
		and isinstance(ip, dpkt.ip6.IP6):
			raise RuntimeError('dpkt parser error: ip4 parsed as ip6')
		
		if eth.type == dpkt.ethernet.ETH_TYPE_IP6 \
		and not isinstance(ip, dpkt.ip6.IP6):
			raise RuntimeError('dpkt parser error: ip6 parsed incorrectly')
		
		if eth.type == dpkt.ethernet.ETH_TYPE_IP:
			if ip.p != dpkt.ip.IP_PROTO_UDP:
				continue

			src_ip = socket.inet_ntoa(ip.src)
			dst_ip = socket.inet_ntoa(ip.dst)
			
		elif eth.type == dpkt.ethernet.ETH_TYPE_IP6:
			if ip.nxt != dpkt.ip.IP_PROTO_UDP:
				continue
			
			src_ip = socket.inet_ntop(socket.AF_INET6, ip.src)
			dst_ip = socket.inet_ntop(socket.AF_INET6, ip.dst)
		
		else:
			continue
		
		udp = ip.data
		try:
			src_port = udp.sport
			dst_port = udp.dport
			payload = udp.data
		except:
			# assume missing UDP header
			src_port = 0
			dst_port = 0
			payload = udp[8:]
		
		yield (ts, ip.src, src_ip, src_port, ip.dst, dst_ip, dst_port, payload)


def transmit(fn, tx_ip, tx_port, fd_data, fd_idx, spd, nskip, ksi, ksp, kdi, kdp, verbose):
	global paused, dropping, t_real, move_up
	
	sck = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
	t_real = time.time()
	t_pcap = None
	
	if fn.endswith('.gz'):
		print('loading gzip', fn)
		pcap_fd = gzip.open(fn)
		fn = fn[:-3]  # get in the car
	else:
		print('loading uncompressed', fn)
		pcap_fd = open(fn, 'rb')
	
	num_seen = 0
	lastprint = 0
	pcap = dpkt.pcapng.Reader(pcap_fd)
	for ts, src_ip_bin, src_ip, src_port, dst_ip_bin, dst_ip, dst_port, payload in get_packets(pcap):
		pkt_len = len(payload)
		if verbose:
			print(src_ip, src_port, dst_ip, dst_port, pkt_len)
		
		if (ksi and ksi != src_ip)   \
		or (ksp and ksp != src_port) \
		or (kdi and kdi != dst_ip)   \
		or (kdp and kdp != dst_port) :
			continue
		
		num_seen += 1
		if nskip > 0:
			nskip -= 1
			continue
		
		try:
			targets[dst_port][0] += 1
			targets[dst_port][1] += pkt_len / 1024.
		except:
			targets[dst_port] = [1, pkt_len / 1024.]
		
		if t_pcap is None:
			t_pcap = ts
			utc = datetime.utcfromtimestamp(ts).strftime('%Y-%m-%d, %H:%M:%S')
			print('first packet was originally sent @{0}, {1}\n'.format(ts, utc))
		
		dt_pcap = ts - t_pcap
		while True:
			t_now = time.time()
			dt_real = t_now - t_real
			if spd == 0:
				break
			dt_comp = dt_real * spd
			if dt_comp >= dt_pcap and paused is None:
				break
			time.sleep(0.005)
		
		if dropping is None:
			dst_ip = tx_ip or dst_ip
			dst_port = tx_port or dst_port
			if fd_data:
				fd_data.write(payload)
				fd_idx.write(
					struct.pack('>QLLHHHH',
						int(ts),                  # uint64 epoch seconds
						(ts-int(ts))*1000000000,  # uint32 epoch decimals
						len(payload),             # uint32 packet length
						len(src_ip_bin),          # uint16 src ip len
						len(dst_ip_bin),          # uint64 dst ip len
						src_port,                 # uint16 src port
						dst_port                  # uint16 dst port
					) + src_ip_bin + dst_ip_bin)
			try:
				sck.sendto(payload, (dst_ip, dst_port))
				#if dst_port == 4321:
				#	with open('dump/{0}-{1}'.format(ts, num_seen), 'wb') as f:
				#		f.write(payload)
			except Exception as e:
				print('\033[1;31m')
				print('error when sending to ip/port [\033[37m{0}\033[31m] [\033[37m{1}\033[31m]\n'.format(dst_ip, dst_port))
				print('if exception is "network unreachable":')
				print('route add -net 224.0.0.0 netmask 224.0.0.0 eth0')
				print('\033[0m')
				raise e
		
		if dropping == 1:
			dropping = None
		
		if t_now > lastprint + 0.05:
			lastprint = t_now
			if move_up > 0:
				print('\033[{0}A'.format(move_up), end='')
			move_up = 3
			print('Playback rate: ', dt_pcap / dt_real, '\033[K')
			print('Total seen :: {0} packets'.format(num_seen))
			for k in targets:
				v = targets[k]
				if v[1] < 1024:
					print('Port {0:5} :: {1} packets, {2:.2f} Kilobyte\033[K'.format(k, v[0], v[1]))
				else:
					print('Port {0:5} :: {1} packets, {2:.2f} Megabyte\033[K'.format(k, v[0], v[1] / 1024))
				move_up += 1
			
			if dropping:
				print('\n** DROPPING **\033[K')
				move_up += 2
			
			print('\033[J')
	
	print('sent', num_seen, 'packets')
	pcap_fd.close()


def main():
	ap = argparse.ArgumentParser(description='replay pcap/pcapng files')
	
	ap.add_argument('files', nargs='+',
		help='list of files to send')

	ap.add_argument('-ksi',  metavar='IP',   default=None,
		help='keep packets with source IP')
	
	ap.add_argument('-ksp',  metavar='PORT', default=None, type=int,
		help='keep packets with source PORT')
	
	ap.add_argument('-kdi',  metavar='IP',   default=None,
		help='keep packets with destination IP')
	
	ap.add_argument('-kdp',  metavar='PORT', default=None, type=int,
		help='keep packets with destination PORT')
	
	ap.add_argument('-di',   metavar='IP',   default=None,
		help='transmit packets to single IP')
	
	ap.add_argument('-dp',   metavar='PORT', default=None, type=int,
		help='transmit packets to single Port')
	
	ap.add_argument('-r',    metavar='RATE', type=int, default=1,
		help='speed multiplier; 1=default, 2=double, 0=turbo')
	
	ap.add_argument('-ss',   metavar='NUM',  type=int, default=0,
		help='skip the first N packets of each pcap file')
	
	ap.add_argument('-o',    metavar='FILE', default=None,
		help='redirect packets to FILE and FILE.idx')
	
	ap.add_argument('-v', action='store_true', help='verbose')
	
	al = ap.parse_args()
	pcap_files = al.files
	tx_ip = al.di
	tx_port = int(al.dp) if al.dp else None
	nskip = al.ss
	spd = al.r
	ksi = al.ksi
	ksp = al.ksp
	kdi = al.kdi
	kdp = al.kdp
	verbose = al.v
	
	fd_data = None
	fd_idx = None
	if al.o:
		fd_data = open(al.o, 'wb')
		fd_idx = open(al.o + '.idx', 'wb')
		fd_idx.write(b'pcpidx\x00\x00')
	
	thr = threading.Thread(target=keyboard)
	thr.daemon = True
	thr.start()

	print('pause and resume:    [p]')
	print('toggle packet drop:  [d]')
	print('drop single packet:  [s]')
	print()
	
	for fn in al.files:
		transmit(fn, tx_ip, tx_port, fd_data, fd_idx, spd, nskip, ksi, ksp, kdi, kdp, verbose)
	
	if fd_data:
		fd_data.close()
		fd_idx.close()


if __name__ == '__main__':
	main()

